"""
Manage the logic of downloading the pokedex and source images.
"""

import re
import json
import asyncio
from pathlib import Path
from dataclasses import dataclass, asdict
from collections import defaultdict

from aiohttp import ClientSession


JS_TO_JSON = re.compile(r"\b([a-zA-Z][a-zA-Z0-9]*?):")


# the dex from showdown assumes only strawberry alcremie, since
# that's what's in showdown, but we might as well add the rest
ALCREMIE_SWEETS = [
  "Strawberry", "Berry", "Love", "Star",
  "Clover", "Flower", "Ribbon",
]

# https://bulbapedia.bulbagarden.net/wiki/List_of_Pok%C3%A9mon_with_gender_differences
# there are some pokemon with notable gender diffs that the dex doesn't cover
# judgement calls made arbitrarily
GENDER_DIFFS = (
  "hippopotas", "hippowdon",
  "unfezant",
  "frillish", "jellicent",
  "pyroar",
  # meowstic, indeedee, basculegion, oinkologne are already handled in the dex
)


@dataclass
class Form:
  name: str
  traits: list[str]
  types: list[str]
  color: str


@dataclass
class Pokemon:
  num: int
  species: str
  forms: list[Form]


async def download_pokedex() -> dict:
  async with ClientSession() as session:
    async with session.get("https://play.pokemonshowdown.com/data/pokedex.js") as res:
      res.raise_for_status()
      text = await res.text("utf-8")
  # this is not json of course, but it's close
  # start by taking out the ; and definition
  cleaned = text.replace("exports.BattlePokedex = ", "").strip(";")
  # then convert the keys to strings
  converted = re.sub(JS_TO_JSON, lambda m: f'"{m.group(1)}":', cleaned)
  # and fix Type: Null, Farfetch'd, Sirfetch'd
  fixed = converted.replace(
    '""Type": Null"', '"Type: Null"').replace("\u2019", "'")
  # then, parse it
  return json.loads(fixed)


def get_traits(species: str, kind: str) -> list[str]:
  traits = []
  if kind in ("mega", "mega-x", "mega-y", "primal"):
    traits.extend(("mega", "nostart"))
  if kind in ("gmax", "eternamax", "rapid-strike-gmax"):
    traits.extend(("gmax", "nostart"))
  if kind in ("alola", "galar", "hisui", "paldea"):
    traits.extend(("regional", kind))

  # special cases
  if species == "Castform" and kind != "base":
    # castform can't start battle in weather forms
    traits.append("nostart")
  if species == "Cherrim" and kind != "base":
    # castform can't start battle in sunshine form
    traits.append("nostart")
  if species == "Aegislash" and kind != "base":
    # aegislash can't start battle in blade form
    traits.append("nostart")
  if species == "Wishiwashi" and kind != "base":
    # wishiwashi can't start battle in schooling form
    traits.append("nostart")
  if species == "Mimikyu" and kind != "base":
    # mimikyu can't start battle in busted form
    traits.append("nostart")
  if species == "Necrozma" and kind == "ultra":
    # necrozma can't start battle in ultra form
    traits.append("nostart")
  if species == "Cramorant" and kind != "base":
    # cramorant can't start battle in gulping/gorging form
    traits.append("nostart")
  if species == "Tauros" and "paldea" in kind:
    # paldean tauros has dumb names
    traits.extend(("regional", "paldea"))
  if species == "Minior" and kind != "meteor":
    # minior can only start the battle in meteor form
    traits.append("nostart")
  if species == "Darmanitan" and "zen" in kind:
    # darmanitan cannot start in zen form
    traits.append("nostart")
    if "galar" in kind:
      # also there's a galar-zen form to handle
      traits.extend(("regional", "galar"))
  if species == "Palafin" and kind == "hero":
    # palafin can only start in zero form
    traits.append("nostart")
  if species == "Gimmighoul" and kind == "roaming":
    # gimmighoul roaming is only in PGO
    traits.append("nostart")

  return sorted(set(traits))


def clean_dex(raw: dict) -> dict[int, Pokemon]:
  regrouped = defaultdict(list)

  for key, entry in raw.items():
    isNonstandard = entry.get("isNonstandard", None)
    baseSpecies = entry.get("baseSpecies", None)
    forme = entry.get("forme", None)
    if isNonstandard not in (None, "Past", "Unobtainable"):
      continue  # remove CAP etc.
    if baseSpecies in ("Pikachu", "Pichu") and forme is not None:
      continue  # remove pikachu spam + spiky ear pichu
    if forme is not None and "Totem" in forme:
      continue  # remove totem pokemon
    if baseSpecies == "Toxtricity" and forme == "Low-Key-Gmax":
      continue  # remove low-key-gmax since it is sort of a duplicate

    num = entry["num"]
    # non-cosmetic forms get separate entries automatically
    # but keeping the separate unown forms would be ridiculous
    if key != "unown" and len(cosmetic := entry.get("cosmeticFormes", [])) > 0:
      cosmetic.append(f'{entry["name"]}-{entry["baseForme"]}')
      if key == "alcremie":
        # oh god this thing
        cosmetic = [
          f"{cf}-{sweet}"
          for cf in cosmetic
          for sweet in ALCREMIE_SWEETS
        ]
      regrouped[num].extend({
        **entry,
        "forme": cf.replace(" ", "-"),
        "formeKind": "cosmetic",
      } for cf in cosmetic)
    elif key in GENDER_DIFFS:
      regrouped[num].append({
        **entry,
        "forme": f'{entry["name"]}-M',
        "formeKind": "cosmetic",
      })
      regrouped[num].append({
        **entry,
        "forme": f'{entry["name"]}-F',
        "formeKind": "cosmetic",
      })
    else:
      regrouped[num].append({
        **entry,
        "forme": entry["name"],
        "formeKind": entry.get("forme", "base").lower(),
      })

  return {
    i: Pokemon(
      num=i,
      species=(
        # doubles as an assertion that forms is not empty
        species := (forms := regrouped[i])[0].get("baseSpecies", forms[0]["name"])
      ),
      forms=[
        Form(
          name=f.get("forme", f["name"]),
          traits=get_traits(species, f["formeKind"].lower()),
          types=f["types"],
          color=f["color"],
        ) for f in forms
      ]
    ) for i in range(1, max(regrouped.keys()) + 1)
  }


async def load_pokedex(dex_file: Path, force_dex: bool) -> dict:
  if dex_file.is_file() and not force_dex:
    with open(dex_file) as infile:
      loaded = json.load(infile)
    dex = {
      int(num): Pokemon(
        num=entry["num"],
        species=entry["species"],
        forms=[Form(**f) for f in entry["forms"]],
      ) for num, entry in loaded.items()
    }
  else:
    # first download the pokedex
    raw_dex = await download_pokedex()
    # clean and reorganize it
    dex = clean_dex(raw_dex)
    # output dex for auditing and reloading
    with open(dex_file, "w") as out:
      json.dump({
        str(i): asdict(pkmn)
        for i, pkmn in dex.items()
      }, out, indent=2)

  return dex


SHOWDOWN_REPLACEMENTS = [
  ("mega-", "mega"),  # charizard, mewtwo
  ("paldea-", "paldea"),  # tauros
  ("mr. ", "mr"),  # mr mime + mr rime
  (" jr.", "jr"),  # mime jr
  ("'d", "d"),  # farfetch'd and sirfetch'd
  ("nidoran-m", "nidoranm"),
  ("-f", "f"),  # gender diff forms
  (re.compile(r"-m$"), ""),  # gender diff forms
  (re.compile(r"^ho-oh$"), "hooh"),
  ("burmy-plant", "burmy"),
  ("shellos-west", "shellos"),
  ("gastrodon-west", "gastrodon"),
  ("hippopotasf", "hippopotas-f"),
  ("hippowdonf", "hippowdon-f"),
  ("rotomf", "rotom-f"),  # rotom fan + frost special case
  ("porygon-z", "porygonz"),
  ("arceusf", "arceus-f"),  # fire, flying, fairy, fighting special case
  ("unfezantf", "unfezant-f"),
  ("-striped", "striped"),  # basculin special cases
  ("-galar-zen", "-galarzen"),  # darmanitan special cases
  ("-spring", ""),  # deerling, sawsbuck special cases
  ("frillishf", "frillish-f"),
  ("jellicentf", "jellicent-f"),
  ("pyroarf", "pyroar-f"),
  ("flabébé", "flabebe"),
  ("flabebe-red", "flabebe"),
  ("floette-red", "floette"),
  ("florges-red", "florges"),
  ("vivillonfancy", "vivillon-fancy"),
  ("vivillon-meadow", "vivillon"),
  ("vivillon-icy-snow", "vivillon-icysnow"),
  ("vivillon-high-plains", "vivillon-highplains"),
  ("meowsticf", "meowstic-f"),
  ("furfrou-la-reine", "furfrou-lareine"),
  ("furfrou-natural", "furfrou"),
  ("%", ""),  # zygarde 10%,
  ("oricorio-pom-pom", "oricorio-pompom"),
  ("oricorio-pa'u", "oricorio-pau"),
  ("type: null", "typenull"),
  ("silvallyf", "silvally-f"),  # fire, flying, fairy, fighting special case
  ("minior-red", "minior"),
  ("mo-o", "moo"),  # jangmo-o, hakamo-o, kommo-o
  ("tapu ", "tapu"),  # tapus
  ("dusk-mane", "duskmane"),  # necrozma
  ("dawn-wings", "dawnwings"),  # necrozma
]


def get_showdown_urls(form: Form) -> list[tuple[str, str]]:
  name = form.name.lower()
  for pat, ins in SHOWDOWN_REPLACEMENTS:
    if isinstance(pat, re.Pattern):
      name = re.sub(pat, ins, name)
    else:
      name = name.replace(pat, ins)
  return [
    (f"https://play.pokemonshowdown.com/sprites/ani/{name}.gif", "gif"),
    (f"https://play.pokemonshowdown.com/sprites/ani-back/{name}.gif", "gif"),
    (f"https://play.pokemonshowdown.com/sprites/gen5/{name}.png", "png"),
    (f"https://play.pokemonshowdown.com/sprites/gen5-back/{name}.png", "png"),
  ]


SEREBII_SPECIAL = {
  "Castform-Rainy": "r",
  "Castform-Snowy": "i",
  "Castform-Sunny": "s",
  "Deoxys-Attack": "a",
  "Deoxys-Defense": "d",
  "Deoxys-Speed": "s",
  "Burmy-Plant": "p",
  "Burmy-Sandy": "s",
  "Burmy-Trash": "t",
  # wormadam plant is default form
  "Wormadam-Sandy": "c",
  "Wormadam-Trash": "t",
  "Cherrim-Sunshine": "s",
  "Shellos-East": "e",
  "Gastrodon-East": "e",
  "Tauros-Paldea-Blaze": "b",
  "Tauros-Paldea-Aqua": "a",
  "Rotom-Heat": "h",
  "Rotom-Wash": "w",
  "Rotom-Frost": "f",
  "Rotom-Fan": "s",
  "Rotom-Mow": "m",
  "Dialga-Origin": "o",
  "Palkia-Origin": "o",
  "Giratina-Origin": "o",
  "Shaymin-Sky": "s",
  "Unfezant-F": "f",
  "Basculin-Blue-Striped": "b",
  "Basculin-White-Striped": "w",
  "Darmanitan-Zen": "z",
  "Darmanitan-Galar-Zen": "gz",
  "Deerling-Summer": "s",
  "Deerling-Autumn": "a",
  "Deerling-Winter": "w",
  "Sawsbuck-Summer": "s",
  "Sawsbuck-Autumn": "a",
  "Sawsbuck-Winter": "w",
  "Frillish-F": "f",
  "Jellicent-F": "f",
  "Kyurem-Black": "b",
  "Kyurem-White": "w",
  "Keldeo-Resolute": "r",
  "Meloetta-Pirouette": "s",
  "Genesect-Douse": "w",
  "Genesect-Shock": "e",
  "Genesect-Burn": "f",
  "Genesect-Chill": "i",
  "Greninja-Ash": "a",
  "Vivillon-Archipelago": "a",
  "Vivillon-Continental": "c",
  "Vivillon-Elegant": "e",
  "Vivillon-Fancy": "f",
  "Vivillon-Garden": "g",
  "Vivillon-High-Plains": "h",
  "Vivillon-Icy-Snow": "i",
  "Vivillon-Jungle": "j",
  "Vivillon-Marine": "ma",
  "Vivillon-Meadow": "m",
  "Vivillon-Modern": "mo",
  "Vivillon-Monsoon": "mon",
  "Vivillon-Ocean": "o",
  "Vivillon-Pokeball": "pb",
  "Vivillon-Polar": "p",
  "Vivillon-River": "r",
  "Vivillon-Sandstorm": "s",
  "Vivillon-Savanna": "sa",
  "Vivillon-Sun": "su",
  "Vivillon-Tundra": "t",
  "Pyroar-F": "f",
  "Flabébé-Blue": "b",
  "Flabébé-Orange": "o",
  "Flabébé-White": "w",
  "Flabébé-Yellow": "y",
  "Flabébé-Red": "r",
  "Floette-Blue": "b",
  "Floette-Orange": "o",
  "Floette-White": "w",
  "Floette-Yellow": "y",
  "Floette-Red": "r",
  "Florges-Blue": "b",
  "Florges-Orange": "o",
  "Florges-White": "w",
  "Florges-Yellow": "y",
  "Florges-Red": "r",
  "Meowstic-F": "f",
  "Furfrou-Dandy": "da",
  "Furfrou-Debutante": "de",
  "Furfrou-Diamond": "d",
  "Furfrou-Heart": "h",
  "Furfrou-Kabuki": "k",
  "Furfrou-La-Reine": "l",
  "Furfrou-Matron": "m",
  "Furfrou-Pharaoh": "p",
  "Furfrou-Star": "s",
  "Aegislash-Blade": "b",
  "Zygarde-10%": "10",
  "Zygarde-Complete": "c",
  "Hoopa-Unbound": "u",
  "Pumpkaboo-Small": "s",  # kinda dumb but w/e
  "Pumpkaboo-Large": "l",
  "Pumpkaboo-Super": "h",
  "Gourgeist-Small": "s",
  "Gourgeist-Large": "l",
  "Gourgeist-Super": "h",
  "Oricorio-Pom-Pom": "p",
  "Oricorio-Pa'u": "pau",
  "Oricorio-Sensu": "s",
  "Lycanroc-Midnight": "m",
  "Lycanroc-Dusk": "d",
  "Wishiwashi-School": "s",
  "Minior-Blue": "b",
  "Minior-Green": "g",
  "Minior-Indigo": "i",
  "Minior-Orange": "o",
  "Minior-Red": "r",
  "Minior-Violet": "v",
  "Minior-Yellow": "y",
  "Mimikyu-Busted": "b",
  "Necrozma-Dusk-Mane": "dm",
  "Necrozma-Dawn-Wings": "dw",
  "Necrozma-Ultra": "m",
  "Magearna-Original": "o",
}
SEREBII_USE_SPECIES = (
  "Shellos-West", "Gastrodon-West",
  "Hippopotas-M", "Hippowdon-M",
  "Unfezant-M",
  "Deerling-Spring", "Sawsbuck-Spring",
  "Frillish-M", "Jellicent-M",
  "Pyroar-M",
  "Furfrou-Natural",
  "Minior-Meteor",
)
SEREBII_IGNORE_MISSING = ("Hippopotas-F", "Hippowdon-F", "Floette-Eternal")


def get_serebii_url(pkmn: Pokemon, form: Form) -> str | None:
  if form.name == pkmn.species or form.name in SEREBII_USE_SPECIES:
    return f"https://www.serebii.net/pokemon/art/{pkmn.num:03d}.png"
  if form.name in SEREBII_SPECIAL:
    return f"https://www.serebii.net/pokemon/art/{pkmn.num:03d}-{SEREBII_SPECIAL[form.name]}.png"
  if "gmax" in form.traits:
    return f"https://www.serebii.net/pokemon/art/{pkmn.num:03d}-gi.png"
  if "mega" in form.traits:
    if "Mega-X" in form.name:
      return f"https://www.serebii.net/pokemon/art/{pkmn.num:03d}-mx.png"
    elif "Mega-Y" in form.name:
      return f"https://www.serebii.net/pokemon/art/{pkmn.num:03d}-my.png"
    else:
      return f"https://www.serebii.net/pokemon/art/{pkmn.num:03d}-m.png"
  if "alola" in form.traits:
    return f"https://www.serebii.net/pokemon/art/{pkmn.num:03d}-a.png"
  if "galar" in form.traits:
    return f"https://www.serebii.net/pokemon/art/{pkmn.num:03d}-g.png"
  if "hisui" in form.traits:
    return f"https://www.serebii.net/pokemon/art/{pkmn.num:03d}-h.png"
  if "paldea" in form.traits:
    return f"https://www.serebii.net/pokemon/art/{pkmn.num:03d}-p.png"
  if pkmn.num in (493, 773):  # arceus and silvally
    type_name = form.name.split('-')[1].lower()
    return f"https://www.serebii.net/pokemon/art/{pkmn.num}-{type_name}.png"
  if "-Therian" in form.name:
    return f"https://www.serebii.net/pokemon/art/{pkmn.num:03d}-s.png"
  if form.name not in SEREBII_IGNORE_MISSING:
    print(f"No Serebii URL known for {form.name}")


async def download(session: ClientSession, url: str, filename: Path) -> tuple[str, Exception | bool]:
  if filename.is_file():
    return url, False
  try:
    async with session.get(url) as res:
      res.raise_for_status()
      with open(filename, "wb") as out:
        out.write(await res.read())
  except Exception as ex:
    return url, ex
  return url, True


async def download_all_for_pokemon(pkmn: Pokemon, image_dir: Path) -> dict[str, dict[str, Exception | bool]]:
  results = defaultdict(dict)
  async with ClientSession() as session:
    for form in pkmn.forms:
      urls = []
      urls += get_showdown_urls(form)
      urls.append((get_serebii_url(pkmn, form), "png"))
      # TODO more sources
      results[form.name].update(await asyncio.gather(*[
        download(session, url, image_dir.joinpath(f"{form.name}-{i}.{ext}"))
        for i, (url, ext) in enumerate(urls) if url is not None
      ]))
  return results


async def download_all(image_dir: Path, pkmn: list[Pokemon]) -> dict[str, dict[str, Exception | bool]]:
  image_dir.mkdir(parents=True, exist_ok=True)
  log = {}
  for p in pkmn:
    log.update(await download_all_for_pokemon(p, image_dir))
  return log


KNOWN_MISSING = [
  "https://play.pokemonshowdown.com/sprites/ani/venusaur-gmax.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/venusaur-gmax.gif",
  "https://play.pokemonshowdown.com/sprites/ani/blastoise-gmax.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/blastoise-gmax.gif",
  "https://play.pokemonshowdown.com/sprites/ani/growlithe-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/growlithe-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani/arcanine-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/arcanine-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani/voltorb-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/voltorb-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani/electrode-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/electrode-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani/tauros-paldeacombat.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/tauros-paldeacombat.gif",
  "https://play.pokemonshowdown.com/sprites/ani/tauros-paldeablaze.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/tauros-paldeablaze.gif",
  "https://play.pokemonshowdown.com/sprites/ani/tauros-paldeaaqua.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/tauros-paldeaaqua.gif",
  "https://play.pokemonshowdown.com/sprites/ani/wooper-paldea.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/wooper-paldea.gif",
  "https://play.pokemonshowdown.com/sprites/ani/qwilfish-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/qwilfish-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani/sneasel-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/sneasel-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani/dialga-origin.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/dialga-origin.gif",
  "https://play.pokemonshowdown.com/sprites/ani/palkia-origin.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/palkia-origin.gif",
  "https://play.pokemonshowdown.com/sprites/ani/lilligant-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/lilligant-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani/basculin-whitestriped.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/basculin-whitestriped.gif",
  "https://play.pokemonshowdown.com/sprites/ani/braviary-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/braviary-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-archipelago.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-archipelago.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-continental.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-continental.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-elegant.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-elegant.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-garden.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-garden.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-highplains.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-highplains.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-icysnow.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-icysnow.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-jungle.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-jungle.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-marine.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-marine.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-modern.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-modern.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-monsoon.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-monsoon.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-ocean.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-ocean.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-polar.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-polar.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-river.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-river.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-sandstorm.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-sandstorm.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-savanna.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-savanna.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-sun.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-sun.png",
  "https://play.pokemonshowdown.com/sprites/gen5/vivillon-tundra.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/vivillon-tundra.png",
  "https://play.pokemonshowdown.com/sprites/gen5/furfrou-dandy.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/furfrou-dandy.png",
  "https://play.pokemonshowdown.com/sprites/gen5/furfrou-debutante.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/furfrou-debutante.png",
  "https://play.pokemonshowdown.com/sprites/gen5/furfrou-diamond.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/furfrou-diamond.png",
  "https://play.pokemonshowdown.com/sprites/gen5/furfrou-heart.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/furfrou-heart.png",
  "https://play.pokemonshowdown.com/sprites/gen5/furfrou-kabuki.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/furfrou-kabuki.png",
  "https://play.pokemonshowdown.com/sprites/gen5/furfrou-lareine.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/furfrou-lareine.png",
  "https://play.pokemonshowdown.com/sprites/gen5/furfrou-matron.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/furfrou-matron.png",
  "https://play.pokemonshowdown.com/sprites/gen5/furfrou-pharaoh.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/furfrou-pharaoh.png",
  "https://play.pokemonshowdown.com/sprites/gen5/furfrou-star.png",
  "https://play.pokemonshowdown.com/sprites/gen5-back/furfrou-star.png",
  "https://play.pokemonshowdown.com/sprites/ani/sliggoo-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/sliggoo-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani/goodra-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/goodra-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani/avalugg-hisui.gif",
  "https://play.pokemonshowdown.com/sprites/ani-back/avalugg-hisui.gif",
]


async def main(
  dex_file: Path, image_dir: Path, startIndex: int, endIndex: int,
  log_skipped: bool, force_dex: bool, dex_only: bool
):
  dex = await load_pokedex(dex_file, force_dex)
  if dex_only:
    return
  log = await download_all(image_dir, (dex[i] for i in range(startIndex, endIndex + 1)))
  new_downloads = 0
  for form, result in log.items():
    for url, info in result.items():
      if isinstance(info, Exception):
        if url not in KNOWN_MISSING:
          print(f"{form}: FAILED {url} - {info}")
      elif not info:
        if log_skipped:
          print(f"{form}: SKIPPED {url} - {info}")
      else:
        print(f"{form}: SUCCESS {url}")
        new_downloads += 1
  print(f"New Downloads: {new_downloads}")


if __name__ == "__main__":
  from argparse import ArgumentParser
  parser = ArgumentParser(
    prog="Image Retriever",
    description="Retrieve pokedex and images",
  )
  parser.add_argument(
    "-d", "--pokedex", default="data/pokedex.json", type=Path, help="Pokedex file"
  )
  parser.add_argument(
    "--refresh-dex", action="store_true", help="Update the pokedex"
  )
  parser.add_argument(
    "--pokedex-only", action="store_true", help="Quit before image download"
  )
  parser.add_argument(
    "-o", "--output", default="images", type=Path, help="Image output directory"
  )
  parser.add_argument(
    "--log-skipped", action="store_true", help="Log skipped images"
  )
  parser.add_argument(
    "bounds", type=lambda a: map(int, a.split("-")), default="1-151", nargs="?",
    help="Range of dex numbers to download, inclusive"
  )

  args = parser.parse_args()
  start, end = args.bounds

  asyncio.run(main(
    args.pokedex, args.output, start, end,
    args.log_skipped, args.refresh_dex, args.pokedex_only
  ))
